from py4j.java_gateway import get_field

import random
import math
import torch
import torch.nn as nn
import torch.optim as optim
from torch.autograd import Variable
import torch.nn.functional as F

from AI_Module import ActionMap
import os
import csv
import copy

class MultiHead(object):
	# hyper parameters
	EPS_START = 0.9  # e-greedy threshold start value
	EPS_END = 0.05  # e-greedy threshold end value
	EPS_DECAY = 10000  # e-greedy threshold decay
	GAMMA = 0.9  # Q-learning discount factor
	LR = 0.001  # NN optimizer learning rate
	INPUT_SIZE = 141 # input layer size
	HIDDEN_SIZE= 80  # NN hidden layer size
	OUTPUT_SIZE = 40  # output layer size
	BATCH_SIZE = 32  # Q-learning batch size
	Memory_Capacity = 100000  # memoory capacity
	UPDATE_TARGET_Q_FREQ = 100 # frequency of updating the target q

	# argument for some utilities
	DirName = "multi-head" # for save
	loadWeightType = "2017AIs" # for loading
	loadAgentType = "ZEN" # for loading
	loadWeightNum = "1250" # for loading
	saveInterval = 50 # Interval of saving the network
	training = 1 # training mode -> 1, acting mode -> 0
	load = 0 # load the weight -> 1, initialize the weight randomely -> 0

	use_cuda = False
	# if gpu is to be used
	# use_cuda = torch.cuda.is_available()
	FloatTensor = torch.cuda.FloatTensor if use_cuda else torch.FloatTensor
	LongTensor = torch.cuda.LongTensor if use_cuda else torch.LongTensor
	ByteTensor = torch.cuda.ByteTensor if use_cuda else torch.ByteTensor
	Tensor = FloatTensor

	def __init__(self, gateway):
		self.gateway = gateway

	def close(self):
		pass

	def getInformation(self, frameData, nonDelay):
		# Load the frame data every time getInformation gets called
		self.frameData = frameData
		self.cc.setFrameData(self.frameData, self.player)
		self.nonDelay = nonDelay
		self.currentFrameNum = nonDelay.getFramesNumber()  # first frame is 14

	# please define this method when you use FightingICE version 3.20 or later
	def roundEnd(self, x, y, z):
		print(x)
		print(y)
		print(z)

		self.isWin()

		if self.training == 1:
			self.saveRoundResult()
			if (self.currentRoundNum % self.saveInterval) == 0:
				self.saveNetwork()

	# please define this method when you use FightingICE version 4.00 or later
	def getScreenData(self, sd):
		pass

	def loadModel(self):
		if self.load == 1:
			# load body 
			self.model.load_state_dict(torch.load("./Weight/"+self.loadWeightType+"/"+self.loadAgentType+"/network/"+self.loadWeightNum+"/body.pth"))
			self.target_model = copy.deepcopy(self.model)
			# load offence Head 
			self.offenceHead.load_state_dict(torch.load("./Weight/"+self.loadWeightType+"/"+self.loadAgentType+"/network/"+self.loadWeightNum+"/offence_head.pth"))
			self.target_offenceHead = copy.deepcopy(self.offenceHead)
			# load defence Head 
			self.defenceHead.load_state_dict(torch.load("./Weight/"+self.loadWeightType+"/"+self.loadAgentType+"/network/"+self.loadWeightNum+"/defence_head.pth"))
			self.target_defenceHead = copy.deepcopy(self.defenceHead)
			print("successfly loaded "+self.loadWeightNum)

	def makeResultFile(self):
		if not os.path.exists("./" + self.DirName + "/network"):
			print("make dir")
			os.makedirs("./"+self.DirName + "/network")

		csvList = []
		csvList.append("roundNum")
		csvList.append("steps")
		csvList.append("averageTotalLoss")
		csvList.append("averageOffenceLoss")
		csvList.append("averageDefenceLoss")
		csvList.append("averageTotalReward")
		csvList.append("averageOffenceReward")
		csvList.append("averageDefenceReward")
		csvList.append("averageOffenceQ")
		csvList.append("averageDefenceQ")
		csvList.append("myHp")
		csvList.append("oppHp")
		csvList.append("score")
		csvList.append("win")
		f = open("./" + self.DirName + "/resultData.csv", 'a')
		writer = csv.writer(f, lineterminator='\n')
		writer.writerow(csvList)
		f.close()

	def initialize(self, gameData, player):
		# Initializng the command center, the simulator and some other things
		self.inputKey = self.gateway.jvm.struct.Key()
		self.frameData = self.gateway.jvm.struct.FrameData()
		self.cc = self.gateway.jvm.aiinterface.CommandCenter()
		self.player = player  # p1 == True, p2 == False
		self.gameData = gameData
		self.simulator = self.gameData.getSimulator()
		self.isGameJustStarted = True

		self.actionMap = ActionMap()
		self.steps_done = 0
		self.currentRoundNum = None

		class ReplayMemory:
			def __init__(self, capacity):
				self.capacity = capacity
				self.memory = []

			def push(self, transition):
				self.memory.append(transition)
				if len(self.memory) > self.capacity:
					del self.memory[0]

			def sample(self, batch_size):
				return random.sample(self.memory, batch_size)

			def __len__(self):
				return len(self.memory)

		class Network(nn.Module):
			def __init__(self, inputSize, hiddenSize, outputSize):
				nn.Module.__init__(self)
				self.l1 = nn.Linear(inputSize, hiddenSize)
				self.l2 = nn.Linear(hiddenSize, hiddenSize)

			def forward(self, x):
				x = F.relu(self.l1(x))
				x = F.relu(self.l2(x))
				return x

		class OffenceHead(nn.Module):
			def __init__(self, hiddenSize, outputSize):
				nn.Module.__init__(self)
				self.l3 = nn.Linear(hiddenSize, outputSize)
			def forward(self, x):
				x = self.l3(x)
				return x

		class DefenceHead(nn.Module):
			def __init__(self, hiddenSize, outputSize):
				nn.Module.__init__(self)
				self.l3 = nn.Linear(hiddenSize, outputSize)
			def forward(self, x):
				x = self.l3(x)
				return x

		# make result file
		self.makeResultFile()

		# memory
		self.memory = ReplayMemory(self.Memory_Capacity)

		# body
		self.model = Network(self.INPUT_SIZE, self.HIDDEN_SIZE, self.OUTPUT_SIZE)
		self.target_model = copy.deepcopy(self.model)
		# offence Head
		self.offenceHead = OffenceHead(self.HIDDEN_SIZE, self.OUTPUT_SIZE)
		self.target_offenceHead = copy.deepcopy(self.offenceHead)
		# defence Head
		self.defenceHead = DefenceHead(self.HIDDEN_SIZE, self.OUTPUT_SIZE)
		self.target_defenceHead = copy.deepcopy(self.defenceHead)
		
		# model load
		self.loadModel()
		
		# optimizer
		self.optimizer = optim.Adam(self.model.parameters(), self.LR)
		self.offenceOptimizer = optim.Adam(self.offenceHead.parameters(), self.LR)
		self.defenceOptimizer = optim.Adam(self.defenceHead.parameters(), self.LR)

		return 0

	def input(self):
		return self.inputKey

	def getObservation(self):
		my = self.frameData.getCharacter(self.player)
		opp = self.frameData.getCharacter(not self.player)

		myHp = abs(my.getHp()/500)
		myEnergy = my.getEnergy()/300
		myX = ((my.getLeft() + my.getRight())/2)/960
		myY = ((my.getBottom() + my.getTop())/2)/640
		mySpeedX = my.getSpeedX()/15
		mySpeedY = my.getSpeedY()/28
		myState = my.getAction().ordinal()

		oppHp = abs(opp.getHp()/500)
		oppEnergy = opp.getEnergy()/300
		oppX = ((opp.getLeft() + opp.getRight())/2)/960
		oppY = ((opp.getBottom() + opp.getTop())/2)/640
		oppSpeedX = opp.getSpeedX()/15
		oppSpeedY = opp.getSpeedY()/28
		oppState = opp.getAction().ordinal()
		oppRemainingFrame = opp.getRemainingFrame()/70

		observation = []
		observation.append(myHp)
		observation.append(myEnergy)
		observation.append(myX)
		observation.append(myY)
		if mySpeedX < 0:
			observation.append(0)
		else:
			observation.append(1)
		observation.append(abs(mySpeedX))
		if mySpeedY < 0:
			observation.append(0)
		else:
			observation.append(1)
		observation.append(abs(mySpeedY))
		for i in range(56):
			if i == myState:
				observation.append(1)
			else:
				observation.append(0)

		observation.append(oppHp)
		observation.append(oppEnergy)
		observation.append(oppX)
		observation.append(oppY)
		if oppSpeedX < 0:
			observation.append(0)
		else:
			observation.append(1)
		observation.append(abs(oppSpeedX))
		if oppSpeedY < 0:
			observation.append(0)
		else:
			observation.append(1)
		observation.append(abs(oppSpeedY))
		for i in range(56):
			if i == oppState:
				observation.append(1)
			else:
				observation.append(0)
		observation.append(oppRemainingFrame)


		myProjectiles = self.frameData.getProjectilesByP1()
		oppProjectiles = self.frameData.getProjectilesByP2()

		if len(myProjectiles) == 2:
			myHitDamage = myProjectiles[0].getHitDamage()/200.0
			myHitAreaNowX = ((myProjectiles[0].getCurrentHitArea().getLeft() + myProjectiles[0].getCurrentHitArea().getRight())/2)/960.0
			myHitAreaNowY = ((myProjectiles[0].getCurrentHitArea().getTop() + myProjectiles[0].getCurrentHitArea().getBottom())/2)/640.0
			observation.append(myHitDamage)
			observation.append(myHitAreaNowX)
			observation.append(myHitAreaNowY)
			myHitDamage = myProjectiles[1].getHitDamage()/200.0
			myHitAreaNowX = ((myProjectiles[1].getCurrentHitArea().getLeft() + myProjectiles[1].getCurrentHitArea().getRight())/2)/960.0
			myHitAreaNowY = ((myProjectiles[1].getCurrentHitArea().getTop() + myProjectiles[1].getCurrentHitArea().getBottom())/2)/640.0
			observation.append(myHitDamage)
			observation.append(myHitAreaNowX)
			observation.append(myHitAreaNowY)
		elif len(myProjectiles) == 1:
			myHitDamage = myProjectiles[0].getHitDamage()/200.0
			myHitAreaNowX = ((myProjectiles[0].getCurrentHitArea().getLeft() + myProjectiles[0].getCurrentHitArea().getRight())/2)/960.0
			myHitAreaNowY = ((myProjectiles[0].getCurrentHitArea().getTop() + myProjectiles[0].getCurrentHitArea().getBottom())/2)/640.0
			observation.append(myHitDamage)
			observation.append(myHitAreaNowX)
			observation.append(myHitAreaNowY)
			for t in range(3):
				observation.append(0.0)
		else:
			for t in range(6):
				observation.append(0.0)

		if len(oppProjectiles) == 2:
			oppHitDamage = oppProjectiles[0].getHitDamage()/200.0
			oppHitAreaNowX = ((oppProjectiles[0].getCurrentHitArea().getLeft() + oppProjectiles[0].getCurrentHitArea().getRight())/2)/960.0
			oppHitAreaNowY = ((oppProjectiles[0].getCurrentHitArea().getTop() + oppProjectiles[0].getCurrentHitArea().getBottom())/2)/640.0
			observation.append(oppHitDamage)
			observation.append(oppHitAreaNowX)
			observation.append(oppHitAreaNowY)
			oppHitDamage = oppProjectiles[1].getHitDamage()/200.0
			oppHitAreaNowX = ((oppProjectiles[1].getCurrentHitArea().getLeft() + oppProjectiles[1].getCurrentHitArea().getRight())/2)/960.0
			oppHitAreaNowY = ((oppProjectiles[1].getCurrentHitArea().getTop() + oppProjectiles[1].getCurrentHitArea().getBottom())/2)/640.0
			observation.append(oppHitDamage)
			observation.append(oppHitAreaNowX)
			observation.append(oppHitAreaNowY)
		elif len(oppProjectiles) == 1:
			oppHitDamage = oppProjectiles[0].getHitDamage()/200.0
			oppHitAreaNowX = ((oppProjectiles[0].getCurrentHitArea().getLeft() + oppProjectiles[0].getCurrentHitArea().getRight())/2)/960.0
			oppHitAreaNowY = ((oppProjectiles[0].getCurrentHitArea().getTop() + oppProjectiles[0].getCurrentHitArea().getBottom())/2)/640.0
			observation.append(oppHitDamage)
			observation.append(oppHitAreaNowX)
			observation.append(oppHitAreaNowY)
			for t in range(3):
				observation.append(0.0)
		else:
			for t in range(6):
				observation.append(0.0)


		# print(len(observation))  #141

		return observation

	def select_action(self, state):
		sample = random.random()
		decay =  (self.EPS_START - self.EPS_END) / self.EPS_DECAY
		eps_threshold = self.EPS_START - (decay * self.steps_done)
		self.steps_done += 1
		if sample > eps_threshold:
			self.model_count_in_a_round += 1 
			# print("model-----------------------")

			offenceQvalues = self.offenceHead(self.model(Variable(state, volatile=True).type(self.FloatTensor))).data
			defenceQvalues = self.defenceHead(self.model(Variable(state, volatile=True).type(self.FloatTensor))).data       
			
			max_totalQ_index = (offenceQvalues + defenceQvalues).max(1)[1].view(1,1)  
			
			max_totalQ = (offenceQvalues + defenceQvalues).max(1)[0][0]
			max_offenceQ = offenceQvalues.max(1)[0][0]
			max_offenceQ_index = offenceQvalues.max(1)[1].view(1,1)
			max_defenceQ = defenceQvalues.max(1)[0][0]
			max_defenceQ_index = defenceQvalues.max(1)[1].view(1,1)
			self.total_offence_Q_in_a_round += max_offenceQ
			self.total_defence_Q_in_a_round += max_defenceQ

			# print("multi max offence Q  : {0}  multi actionName : {1}".format(max_offenceQ, self.actionMap.actionMap[int(max_offenceQ_index[0,0])]))   
			# print("multi max defence Q : {0}  multi actionName : {1}".format(max_defenceQ, self.actionMap.actionMap[int(max_defenceQ_index[0,0])]))
			# print("multi max total Q    : {0}  multi actionName : {1}".format(max_totalQ, self.actionMap.actionMap[int(max_totalQ_index[0,0])]))

			return max_totalQ_index
		else:
			# print("random-----------------------")
			return self.LongTensor([[random.randrange(40)]])

	def select_maxAction(self, state):
		# print("multi select max q value-----------------------")
		offenceQvalues = self.offenceHead(self.model(Variable(state, volatile=True).type(self.FloatTensor))).data
		defenceQvalues = self.defenceHead(self.model(Variable(state, volatile=True).type(self.FloatTensor))).data       
		max_totalQ_index = (offenceQvalues + defenceQvalues).max(1)[1].view(1,1)  

		return max_totalQ_index
			

	# return offenceReward ,defenceReward
	def makeReward(self):
		self.reward_count_in_a_round += 1
		offenceReward = self.oppLastHp - self.nonDelay.getCharacter(not self.player).getHp()
		defenceReward = self.nonDelay.getCharacter(self.player).getHp() - self.myLastHp 
		# print("offence reward : {0}  defencer reward : {1}".format(offenceReward, defenceReward))
		self.total_offence_reward_in_a_round += offenceReward
		self.total_defence_reward_in_a_round += defenceReward
		# print("offence: ", offenceReward)
		# print("defence: ", defenceReward)
		return offenceReward, defenceReward
						
	def setLastHp(self):
		self.myLastHp = self.nonDelay.getCharacter(self.player).getHp()
		self.oppLastHp = self.nonDelay.getCharacter(not self.player).getHp()

	def ableAction(self):
		if self.nonDelay.getCharacter(self.player).isControl() == True:
			return True
		else:
			return False

	def playAction(self, actionNum):
		actionName = self.actionMap.actionMap[int(actionNum)]
		self.cc.commandCall(actionName)

	def pushData(self, next_state, offenceReward, defenceReward):
		self.memory.push((self.FloatTensor([self.state]), self.action, self.FloatTensor([next_state]), self.FloatTensor([offenceReward]), self.FloatTensor([defenceReward])))

	def learn(self):
		if len(self.memory) < self.BATCH_SIZE:
			return
		self.steps_learn_in_a_round += 1

		transitions = self.memory.sample(self.BATCH_SIZE)

		batch_state, batch_action, batch_next_state, batch_offence_reward, batch_defence_reward = zip(*transitions)

		batch_state = Variable(torch.cat(batch_state))
		batch_action = Variable(torch.cat(batch_action))
		batch_offence_reward = Variable(torch.cat(batch_offence_reward))
		batch_defence_reward = Variable(torch.cat(batch_defence_reward))
		batch_next_state = Variable(torch.cat(batch_next_state))
		# current Q values are estimated by NN for all actions
		current_offence_q_values = self.offenceHead(self.model(batch_state)).gather(1, batch_action)
		current_defence_q_values = self.defenceHead(self.model(batch_state)).gather(1, batch_action)
		# expected Q values are estimated from actions which gives maximum Q value
		max_next_offence_q_values = self.target_offenceHead(self.target_model(batch_next_state)).detach().max(1)[0]
		max_next_defence_q_values = self.target_defenceHead(self.target_model(batch_next_state)).detach().max(1)[0]
		expected_offence_q_values = batch_offence_reward + (self.GAMMA * max_next_offence_q_values)
		expected_defence_q_values = batch_defence_reward + (self.GAMMA * max_next_defence_q_values)
		# loss is measured from error between current and newly expected Q values
		offenceLoss = F.smooth_l1_loss(current_offence_q_values, expected_offence_q_values)
		defenceLoss = F.smooth_l1_loss(current_defence_q_values, expected_defence_q_values)
		all_Loss = offenceLoss + defenceLoss
		# print("all_loss : {}".format(all_Loss.data[0]))
		# print("offence_loss : {}".format(offenceLoss.data[0]))
		# print("defence_loss : {}".format(defenceLoss.data[0]))
		self.total_loss_in_a_round += all_Loss.data[0]
		self.total_offence_loss_in_a_round += offenceLoss.data[0]
		self.total_defence_loss_in_a_round += defenceLoss.data[0]

		self.optimizer.zero_grad()
		self.offenceOptimizer.zero_grad()
		self.defenceOptimizer.zero_grad()
		all_Loss.backward()
		self.optimizer.step()
		self.offenceOptimizer.step()
		self.defenceOptimizer.step()

		# target Q
		if self.steps_learn_in_a_round % self.UPDATE_TARGET_Q_FREQ == 0:
			self.target_model = copy.deepcopy(self.model)
			self.target_offenceHead = copy.deepcopy(self.offenceHead)
			self.target_defenceHead = copy.deepcopy(self.defenceHead)

	def isWin(self):
		if self.nonDelay.getCharacter(self.player).getHp() == self.nonDelay.getCharacter(not self.player).getHp():
			if self.training == 1:
				self.win = 0
				print("multi win")
			else:
				self.win = 1
				print("multi lose")
		elif self.nonDelay.getCharacter(self.player).getHp() > self.nonDelay.getCharacter(not self.player).getHp():
			self.win = 1
			print("multi win")
		else:
			self.win = 0
			print("multi lose")

	def saveRoundResult(self):
		if self.nonDelay.getCharacter(not self.player).getHp() == 0:
			score = 0
		else:
			score = (self.nonDelay.getCharacter(not self.player).getHp()/(self.nonDelay.getCharacter(not self.player).getHp() + self.nonDelay.getCharacter(self.player).getHp())) * 1000
		if self.steps_learn_in_a_round == 0:
			average_total_loss = 0
			average_offence_loss = 0
			average_defence_loss = 0
		else:
			average_total_loss = self.total_loss_in_a_round/self.steps_learn_in_a_round
			average_offence_loss = self.total_offence_loss_in_a_round/self.steps_learn_in_a_round
			average_defence_loss = self.total_defence_loss_in_a_round/self.steps_learn_in_a_round
		if self.model_count_in_a_round == 0:
			average_offence_Q = 0
			average_defence_Q = 0
		else:
			average_offence_Q = self.total_offence_Q_in_a_round/self.model_count_in_a_round
			average_defence_Q = self.total_defence_Q_in_a_round/self.model_count_in_a_round
		average_offence_reward = self.total_offence_reward_in_a_round/self.reward_count_in_a_round
		average_defence_reward = self.total_defence_reward_in_a_round/self.reward_count_in_a_round
		average_total_reward = (self.total_offence_reward_in_a_round+self.total_defence_reward_in_a_round)/self.reward_count_in_a_round

		csvList = []
		csvList.append(self.currentRoundNum)
		csvList.append(self.steps_done)

		csvList.append(average_total_loss)
		csvList.append(average_offence_loss)
		csvList.append(average_defence_loss)
		csvList.append(average_total_reward)
		csvList.append(average_offence_reward)
		csvList.append(average_defence_reward)
		csvList.append(average_offence_Q)
		csvList.append(average_defence_Q)
		
		csvList.append(abs(self.nonDelay.getCharacter(self.player).getHp()))
		csvList.append(abs(self.nonDelay.getCharacter(not self.player).getHp()))
		csvList.append(score)
		csvList.append(self.win)
		print("round num : {}".format(self.currentRoundNum))
		print("average_total_loss : {}".format(average_total_loss))
		print("average_offence_loss : {}".format(average_offence_loss))
		print("average_defence_loss : {}".format(average_defence_loss))
		print("average_total_reward : {}".format(average_total_reward))
		print("average_offence_reward : {}".format(average_offence_reward))
		print("average_defence_reward : {}".format(average_defence_reward))
		print("average_offence_Q : {}".format(average_offence_Q))
		print("average_defence_Q : {}".format(average_defence_Q))

		f = open("./" + self.DirName + "/resultData.csv", 'a')
		writer = csv.writer(f, lineterminator='\n')
		writer.writerow(csvList)
		f.close()

		print("multi-head save the round result data")

	def saveNetwork(self):
		if not os.path.exists("./" + self.DirName + "/network/" + str(self.currentRoundNum)):
			os.makedirs("./"+self.DirName + "/network/" + str(self.currentRoundNum))
		torch.save(self.model.state_dict(), "./" + self.DirName + "/network/" + str(self.currentRoundNum) + "/body.pth")
		torch.save(self.offenceHead.state_dict(), "./" + self.DirName + "/network/" + str(self.currentRoundNum) + "/offence_head.pth")
		torch.save(self.defenceHead.state_dict(), "./" + self.DirName + "/network/" + str(self.currentRoundNum) + "/defence_head.pth")
		print("save the network")

	def processing(self):
		# First we check whether we are at the end of the round
		if self.frameData.getEmptyFlag() or self.frameData.getRemainingFramesNumber() <= 0:
			self.isGameJustStarted = True
			return
		if not self.isGameJustStarted:
			pass
		else:
			# this "else" is going to be called in the first frame in the round
			self.isGameJustStarted = False
			self.currentRoundNum = self.frameData.getRound()
			self.steps_learn_in_a_round = 0
			self.total_loss_in_a_round = 0
			self.model_count_in_a_round = 0
			self.total_offence_Q_in_a_round = 0
			self.total_defence_Q_in_a_round = 0
			self.total_offence_loss_in_a_round = 0
			self.total_defence_loss_in_a_round = 0
			self.reward_count_in_a_round = 0
			self.total_offence_reward_in_a_round = 0
			self.total_defence_reward_in_a_round = 0

		if self.cc.getSkillFlag():
			self.inputKey = self.cc.getSkillKey()
			return
		self.inputKey.empty()
		self.cc.skillCancel()

		if self.training == 1:
			# this "if" is going to be called in the first frame in the round
			if self.currentFrameNum == 14:
				print("first")
				self.state = self.getObservation()
				self.setLastHp()
				action = self.select_action(self.FloatTensor([self.state]))
				self.action = action
				self.playAction(action[0, 0])
			
			elif self.ableAction():
				offenceReward, defenceReward = self.makeReward()
				state = self.getObservation()
				self.pushData(state, offenceReward, defenceReward)
				self.learn()
				self.setLastHp()
				self.state = state
				action = self.select_action(self.FloatTensor([state]))
				self.action = action
				self.playAction(action[0, 0])

		else:
			if self.ableAction():
				state = self.getObservation()
				action = self.select_maxAction(self.FloatTensor([state]))
				self.playAction(action[0, 0])


	class Java:
		implements = ["aiinterface.AIInterface"]













